package main

import (
	"context"
	"encoding/json"
	"log"
	"sync"
	"time"

	"git.joco.dk/sng/fermentord/internal/configuration"
	"git.joco.dk/sng/fermentord/internal/controllers"
	"git.joco.dk/sng/fermentord/internal/dwingest"
	"git.joco.dk/sng/fermentord/internal/hw"
	"git.joco.dk/sng/fermentord/internal/lcd"
	"git.joco.dk/sng/fermentord/pkg/temperature"
	"git.joco.dk/sng/fermentord/pkg/tilt"
	"github.com/fsnotify/fsnotify"
	"github.com/getsentry/sentry-go"
	"github.com/nats-io/nats.go"
	"github.com/spf13/viper"
)

func mainLoop(ctx context.Context, wg *sync.WaitGroup, js nats.JetStream, config *configuration.Configuration) {
	hub := sentry.CurrentHub().Clone()
	defer hub.Flush(10 * time.Second)
	defer wg.Done()

	// Display
	display, err := lcd.NewLCD()
	if err != nil {
		hub.CaptureException(err)
		log.Fatal(err)
	}
	defer display.Close()

	// Controller
	ctrl := controllers.NewChamberController("Chamber 1", *config)

	// Configuration reload
	loadConfiguration := func() {
		temperature.ConfigUpdate <- *config
		ctrl.ConfigUpdates <- *config
		display.SetSetpointTemp(config.FermentationTemperature)
	}

	viper.OnConfigChange(func(in fsnotify.Event) {
		loadConfiguration()
	})
	loadConfiguration()
	viper.WatchConfig()

	// NATS
	ingest := dwingest.NewDWIngest()

	gpio, err := hw.NewGpio()
	if err != nil {
		hub.CaptureException(err)
		log.Fatal(err)
	}
	defer gpio.Close()

	wg.Add(5)
	go display.Run(ctx, wg)
	go ctrl.Run(ctx, wg)
	go ingest.Run(ctx, wg, js, config)
	go temperature.PollSensors(ctx, wg, 1*time.Second, config.Sensors.Weight)
	go tilt.PollSensors(ctx, wg, 1*time.Minute, 20*time.Second)

	for {
		select {
		case reading := <-temperature.Reading:
			ctrl.SetTemperature(reading)
			ingest.AddReading(reading)
			display.SetTemperature(reading)

		case ev := <-gpio.C:
			var evs string

			switch ev {
			case hw.DoorClosed:
				gpio.LightsOff()
				gpio.StartFan()
				ctrl.Resume()
				evs = "DOOR_CLOSED"

			case hw.DoorOpened:
				ctrl.Pause()
				gpio.LightsOn()
				gpio.StopFan()
				evs = "DOOR_OPENED"
			}

			b, err := json.Marshal(map[string]interface{}{
				"time":  time.Now().UTC(),
				"event": evs,
			})
			if err != nil {
				hub.CaptureException(err)
				log.Printf("Error marshaling JSON: %v", err)
				break
			}

			// Publish to NATS
			_, err = js.Publish(config.NATS.Subject.Event, b)
			if err != nil {
				hub.CaptureException(err)
				log.Printf("Error publishing to NATS: %v", err)
			}

		case state := <-ctrl.C:
			gpioSetState(state, gpio)
			ingest.AddState(state)
			display.SetState(state)

		case <-ctx.Done():
			return
		}
	}
}
